﻿using System;
using System.Globalization;
using System.Collections.Generic;
using System.IO;

namespace TouchTokens
{
	/*
	 *   Authors: Caroline Appert (caroline.appert@lri.fr)
	 *   Copyright (c) Univ. Paris-Sud XI, 2015. All Rights Reserved
	 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
	*/

	public class TouchTokenRecognizer
	{

		public List<TokenTemplate> tokenTemplates;

		public TouchTokenRecognizer(String path)
		{
			this.tokenTemplates = new List<TokenTemplate>();
			try
			{
				StreamReader br = new StreamReader(path);
				string line = br.ReadLine();
				//			NumberFormat format = NumberFormat.getInstance(Locale.US);
				NumberFormatInfo df = new NumberFormatInfo();
				//DecimalFormatSymbols symbols = new DecimalFormatSymbols();
				df.NumberDecimalSeparator = ".";
				df.NumberGroupSeparator = ",";
				//df.DecimalFormatSymbols = symbols;
				while (!string.ReferenceEquals(line, null))
				{
					
					if (!line.StartsWith("#", StringComparison.Ordinal))
					{
						//Console.WriteLine(""+line);
						string[] parts = line.Split(',');
						if (parts.Length > 3)
						{
							string tokenID = parts[0];
							int fingerCount = int.Parse(parts[1]);
							List<TouchPoint> points = new List<TouchPoint>();
							for (int i = 0; i < fingerCount; i++)
							{
								double x = Double.Parse(parts[2 + (i * 2)], df);
								double y = Double.Parse(parts[2 + (i * 2) + 1], df);
								//Console.WriteLine("coordonnées point " + i + " : " + x + " " + y);
								TouchPoint point = new TouchPoint(x, y);
								points.Add(point);
							}
							TokenTemplate tt = new TokenTemplate(points, tokenID);
							//						System.WriteLine(tt);
							tokenTemplates.Add(tt);
							int index = (fingerCount-1)*2 + 4;
							if (parts.Length >= index + 4){
								int[] rgba = new int[4];
								rgba[0] = int.Parse(parts[index]);
								rgba[1] = int.Parse(parts[index+1]);
								rgba[2] = int.Parse(parts[index+2]);
								rgba[3] = int.Parse(parts[index+3]);
								tt.RGBA = rgba;
							}
							index = index + 4;
							if (parts.Length >= index + 1){
								tt.Shape = parts[index];
							}
							if (parts.Length >= index + 2){
								tt.Ray = Double.Parse(parts[index+1]);
							}
							if (parts.Length >= index + 3){
								tt.Inner = Double.Parse(parts[index+2]);
							}
						}

					}
					line = br.ReadLine();
				}
				br.Close();
			}
			catch (FileNotFoundException e)
			{
				Console.WriteLine(e.ToString());
				Console.Write(e.StackTrace);
			}
			catch (IOException e)
			{
				Console.WriteLine(e.ToString());
				Console.Write(e.StackTrace);
			}
			catch (FormatException e)
			{
				Console.WriteLine(e.ToString());
				Console.Write(e.StackTrace);
			}

		}

		//Set the refPoint as the new origin for points
		public static void centerOnPoint(List<TouchPoint> points, TouchPoint refPoint)
		{
			foreach (TouchPoint pt in points)
			{
				pt.setLocation(pt.X - refPoint.X, pt.Y - refPoint.Y);
			}
		}

		//Calc the angle between 2 vectors
		public static double angleBetweenVectors(TouchPoint vector1, TouchPoint vector2)
		{
			double angle = Math.Atan2(vector2.Y, vector2.X) - Math.Atan2(vector1.Y, vector1.X);
			//	normalize to the range 0 .. 2 * Pi:
			if (angle < 0)
			{
				angle += 2 * Math.PI;
			}
			return angle;
		}

		//Sort the points CCW according to refPoint
		public static void orderCCW(List<TouchPoint> points, TouchPoint refPoint)
		{
			TouchPoint vectorXAxis = new TouchPoint(10, 0);
			points.Sort(new ComparatorAnonymousInnerClass(refPoint, vectorXAxis));
		}

		private class ComparatorAnonymousInnerClass : IComparer<TouchPoint>
		{
			private TouchPoint refPoint;
			private TouchPoint vectorXAxis;

			public ComparatorAnonymousInnerClass(TouchPoint refPoint, TouchPoint vectorXAxis)
			{
				this.refPoint = refPoint;
				this.vectorXAxis = vectorXAxis;
			}

			public int Compare(TouchPoint pt1, TouchPoint pt2)
			{
				TouchPoint vector1 = new TouchPoint(pt1.X - refPoint.X, pt1.Y - refPoint.Y);
				double angle1 = angleBetweenVectors(vectorXAxis, vector1);
				TouchPoint vector2 = new TouchPoint(pt2.X - refPoint.X, pt2.Y - refPoint.Y);
				double angle2 = angleBetweenVectors(vectorXAxis, vector2);
				return (int)((angle2 - angle1) * 100);
			}
		}

		public static void rotateBy(List<TouchPoint> points, TouchPoint refPoint, double theta, List<TouchPoint> newPoints)
		{
			TouchPoint ptSrc, ptDest;
			for (int i = 0; i < points.Count; i++)
			{
				ptSrc = points[i];
				if (newPoints.Count > i)
				{
					ptDest = newPoints[i];
				}
				else
				{
					ptDest = new TouchPoint();
					newPoints.Insert(i, ptDest);
				}
				double x = (ptSrc.X - refPoint.X) * Math.Cos(theta) - (ptSrc.Y - refPoint.Y) * Math.Sin(theta) + refPoint.X;
				double y = (ptSrc.X - refPoint.X) * Math.Sin(theta) + (ptSrc.Y - refPoint.Y) * Math.Cos(theta) + refPoint.Y;
				ptDest.setLocation(x, y);
			}
		}

		public static List<TouchPoint> orderAndAlign(List<TouchPoint> points, TouchPoint refPoint)
		{
			List<TouchPoint> resPoints = new List<TouchPoint>();
			foreach (TouchPoint pt in points)
			{
				resPoints.Add(new TouchPoint(pt.X, pt.Y));
			}
			centerOnPoint(resPoints, refPoint);
			TouchPoint origin = new TouchPoint(0, 0);
			orderCCW(resPoints, origin);
			TouchPoint refVector = new TouchPoint(resPoints[0].X, resPoints[0].Y);
			TouchPoint xAxisVector = new TouchPoint(10, 0);
			double angle = angleBetweenVectors(refVector, xAxisVector);
			rotateBy(resPoints, origin, angle, resPoints);
			return resPoints;
		}

		//Find the triangle centroid
		public static TouchPoint centroid(List<TouchPoint> points)
		{
			double sumX = 0;
			double sumY = 0;
			for (IEnumerator<TouchPoint> iterator = points.GetEnumerator(); iterator.MoveNext();)
			{
				TouchPoint next = iterator.Current;
				sumX += next.X;
				sumY += next.Y;
			}
			int length = points.Count;
			return new TouchPoint(sumX / length, sumY / length);
		}

		//Return a list of all the different permutations for a token
		public static List<List<TouchPoint>> allPermutations(List<TouchPoint> points)
		{
			List<List<TouchPoint>> allPermutations = new List<List<TouchPoint>>();
			for (int i = 0; i < points.Count; i++)
			{
				List<TouchPoint> order = new List<TouchPoint>();
				order.Add(points[i]);
				for (int j = i + 1; j < points.Count; j++)
				{
					order.Add(points[j]);
				}
				for (int j = 0; j < i; j++)
				{
					order.Add(points[j]);
				}
				allPermutations.Add(order);
			}
			return allPermutations;
		}

		//...
		public static double meanSquaredPairedPointDistanceForAlignedPairs(List<TouchPoint> points1, List<TouchPoint> points2)
		{
			if (points1.Count != points2.Count)
			{
				return double.MaxValue;
			}
			else
			{
				double res = 0;
				for (int i = 0; i < points1.Count; i++)
				{
					TouchPoint p1 = points1[i];
					TouchPoint p2 = points2[i];
					res += (p1.X - p2.X) * (p1.X - p2.X) + (p1.Y - p2.Y) * (p1.Y - p2.Y);
				}
				return res / points1.Count;
			}
		}

		public static double minDistance(List<TouchPoint> templatePointsOrderedAndAligned, List<TouchPoint> inputPoints, TouchPoint refPoint)
		{
			List<TouchPoint> inputPointsOrderedAndAligned = orderAndAlign(inputPoints, refPoint);
			List<List<TouchPoint>> allPermutation = allPermutations(inputPointsOrderedAndAligned);
			double minDis = double.MaxValue;
			foreach (List<TouchPoint> inputPermutation in allPermutation)
			{
				TouchPoint origin = new TouchPoint(0, 0);
				TouchPoint refVector = new TouchPoint(inputPermutation[0].X, inputPermutation[0].Y);
				TouchPoint xAxisVector = new TouchPoint(10, 0);
				double angle = angleBetweenVectors(refVector, xAxisVector);
				List<TouchPoint> inputPermutationRotated = new List<TouchPoint>();
				rotateBy(inputPermutation, origin, angle, inputPermutationRotated);
				double d = meanSquaredPairedPointDistanceForAlignedPairs(inputPermutationRotated, templatePointsOrderedAndAligned);
				if (d < minDis)
				{
					minDis = d;
				}
			}
			return minDis;
		}

		public virtual TokenTemplate recognize(List<TouchPoint> input)
		{
			//Debug.Log("Nb de templates : " + tokenTemplates.Count);
			foreach (TokenTemplate tokenTemplate in tokenTemplates)
			{
				tokenTemplate.Distance = double.MaxValue;
			}
			double minDis = double.MaxValue;

			TokenTemplate tokenRecognized = null;
			foreach (TokenTemplate tokenTemplate in tokenTemplates)
			{
				List<TouchPoint> templatePoints = orderAndAlign(tokenTemplate.Points, centroid(tokenTemplate.Points));
				double d = minDistance(templatePoints, input, centroid(input));
				//Debug.Log("distance to " + tokenTemplate.tokenID + " : " + d);
				tokenTemplate.Distance = d;
				//			System.out.println("\t--> distance to "+tokenTemplate+" is "+tokenTemplate.distance);
				if (d < minDis)
				{
					//Debug.Log("d<minDis : " + (d < minDis));
					minDis = d;
					tokenRecognized = tokenTemplate;
				}
		
			}
			
			return tokenRecognized;
		}

    }

}
